package com.edtoon.rott;

import com.edtoon.rott.ott.OTTConstants;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;

/*
 * See http://www.logomanager.co.uk/help/Edit/RTX.html for RTX spec.
 * See http://members.lycos.nl/jupp/linux/soft/rtttl_player/EBNF.txt for RTTTL spec.
 */
public class ToneParser
{
	private static final boolean DEBUG = false;

	private int defaultDuration = OTTConstants.DEFAULT_DURATION, defaultScale = OTTConstants.DEFAULT_SCALE,
		defaultBPM = OTTConstants.DEFAULT_BPM, controlLoop = 0, elementCount = 0;
	private StringBuffer buffer = new StringBuffer();
	private ArrayList notes = new ArrayList();
	private OutputStream oStream = null;
	private boolean doOutput = false;

	public ToneParser() { }

	public void setDoOutput(final boolean bool)
	{
		doOutput = bool;
	}

	private static void debug(final String str)
	{
		if(DEBUG)
		{
			System.out.println(str);
		}
	}

	private void output(final String str)
	{
		if(doOutput)
		{
			System.out.println(str);
		}
	}

	public void parse(final byte[] bytes)
		throws ToneParseException, IOException
	{
		parse(bytes, null);
	}

	public void parse(final byte[] bytes, final OutputStream os)
		throws ToneParseException, IOException
	{
		String rtxData = new String(bytes), name = null, controlSection = null, toneSection = null;
		StringTokenizer sections = new StringTokenizer(rtxData, ":");

		if(sections.countTokens() != 3)
		{
			throw new ToneParseException(Errors.ERROR_RTX_TOO_MANY_SECTIONS,
				"Input contained too many sections");
		}

		oStream = os;

		name = sections.nextToken().trim();
		controlSection = sections.nextToken().trim().toLowerCase();
		toneSection = sections.nextToken().trim().toLowerCase();

		if(name.length() > 5)
		{
			name = name.substring(0, 10);
			debug("Song name chopped to: " + name);
		} else {
			output("Song name: " + name);
		}

		parseControlSection(controlSection);
		parseToneSection(toneSection);

		buffer.append("00000010"); // command length
		debug("-8b- Command length: 2");
		buffer.append("01001010"); // ringing tone programming
		debug("-8b- Ringing tone programming");
		buffer.append("0011101"); // sound
		debug("-7b- Sound");
		buffer.append("001"); // basic song type
		debug("-3b- Basic song type");

		int len = name.length();

		buffer.append(getBits(len, 4));
		debug("-4b- Title length: " + len);

		for(int i = 0; i < len; i++)
		{
			buffer.append(getBits(name.charAt(i), 8));
		}

		debug("-" + (name.length()*8) + "b- Title: " + name);

		buffer.append("00000001"); // song sequence length
		debug("-8b- Song sequence length: 1");
		buffer.append("000"); // pattern header id
		debug("-3b- Pattern Header ID");
		buffer.append("00"); // pattern type
		debug("-2b- Pattern ID: A part");
		buffer.append(getBits(controlLoop, 4)); // loop value
		debug("-4b- Loop value: " + controlLoop);

		buffer.append(getBits(((elementCount*2)+2), 8)); // length of pattern instruction
		debug("-8b- Pattern length: " + ((elementCount*2)+2));
		// for some reason MobileStreams adds one to pattern length though the last pattern does not exist
		int count = 1;

		buffer.append(OTTConstants.TEMPO_INSTRUCTION_ID); // tempo instruction id
		buffer.append(getTempoString(defaultBPM));
		debug("-8b- [" + count + "] Tempo instruction: " + defaultBPM);

		for(final Iterator i = notes.iterator(); i.hasNext(); )
		{
			Note note = (Note)i.next();
			int noteID = 0;

			count++;

			buffer.append(OTTConstants.SCALE_INSTRUCTION_ID); // scale instruction id
			buffer.append(getScaleString(note.scale));
			debug("-5b- [" + count + "] Scale instruction: " + note.scale);

			count++;

			buffer.append(OTTConstants.NOTE_INSTRUCTION_ID); // note instruction id
			buffer.append((String)OTTConstants.noteMap.get(note.note));
			buffer.append(getDurationString(note.duration));

			if(note.dotted)
			{
				buffer.append("01"); // note duration specifier
			} else {
				buffer.append("00"); // note duration specifier
			}

			debug("-12- [" + count + "] Note instruction: " + note.note + ", duration: " + note.duration +
				", special: " + note.dotted);
		}

		int mod = buffer.length() % 8;

		if(mod != 0)
		{
			buffer.append(getBits(0, 8 - mod));
			debug("Padding to close: " + getBits(0, 8 - mod));
		} else {
			debug("NOT PADDING TO CLOSE");
		}

		buffer.append(OTTConstants.COMMAND_END); // command end
		debug("-8b- Command end");

		String bufResult = buffer.toString();
		int dataLength = bufResult.length() / 8;
		byte[] data = new byte[dataLength];

		for(int i = 0; i < dataLength; i++)
		{
			int result = Integer.parseInt(bufResult.substring(i * 8, i * 8 + 8), 2);
			data[i] = (byte)result;
		}

		if(oStream != null)
		{
			oStream.write(data);
		}
	}

	private void parseNote(final String str, int duration)
		throws ToneParseException
	{
		int strlen = str.length(), scale = defaultScale;
		String text = text = "Note: ";
		char ch = (char)-1;
		String note = null;
		boolean dotted = false;

		for(int i = 0; i < strlen; i++)
		{
			ch = str.charAt(i);

			if(i == 0)
			{
				if(ch == 'p' || ch == 'e' || ch == 'h' || ch == 'b')
				{
					note = "" + ch;
				} else if(ch == 'c' || ch == 'd' || ch == 'f' || ch == 'g' || ch == 'a') {
					if(strlen > 1 && str.charAt(1) == '#')
					{
						note = "" + ch + "#";
						i++;
					} else {
						note = "" + ch;
					}
				} else {
					throw new ToneParseException(Errors.ERROR_RTX_INVALID_NOTE_FORMAT,
						"Invalid format for a note");
				}
			} else if(ch == '.') {
				dotted = true;
			} else if(ch == '4') {
				scale = 4;
			} else if(ch == '5') {
				scale = 5;
			} else if(ch == '6') {
				scale = 6;
			} else if(ch == '7') {
				scale = 7;
			} else {
				throw new ToneParseException(Errors.ERROR_RTX_INVALID_NOTE_FORMAT,
					"Invalid format for a note");
			}
		}

		if(dotted)
		{
			text += "Dotted ";
		}

		text += note;
		text += ", Duration: " + duration;
		text += ", Scale: " + scale;
		text += ", Tempo: " + defaultBPM;

		output(text);

		Note noteObj = new Note();
		noteObj.scale = scale;
		noteObj.duration = duration;
		noteObj.dotted = dotted;
		noteObj.note = note;
		noteObj.bpm = defaultBPM;
		notes.add(noteObj);
		elementCount++;
	}

	private void parseTone(final String str)
		throws ToneParseException
	{
		int strlen = str.length();
		char ch = (char)-1;
		int duration = defaultDuration;

		for(int i = 0; i < strlen; i++)
		{
			ch = str.charAt(i);

			if(i == 0)
			{
				if(ch == '2')
				{
					duration = 2;
				} else if(ch == '4') {
					duration = 4;
				} else if(ch == '8') {
					duration = 8;
				} else if(ch == '1') {
					if(strlen > 1 && str.charAt(1) == '6')
					{
						i++;
						duration = 16;
					} else {
						duration = 1;
					}
				} else if(ch == '3') {
					if(strlen > 1 && str.charAt(1) == '2')
					{
						i++;
						duration = 32;
					} else {
						throw new ToneParseException(Errors.ERROR_RTX_INVALID_SCALE_SETTING,
							"Invalid scale setting for a note");
					}
				} else {
					parseNote(str.substring(i), duration);
					break;
				}
			} else {
				parseNote(str.substring(i), duration);
				break;
			}
		}
	}

	private void parseToneSection(final String str)
		throws ToneParseException
	{
		StringTokenizer parts = new StringTokenizer(str, ","), control = null;
		String tmp = null;

		while(parts.hasMoreTokens())
		{
			tmp = parts.nextToken().trim();

			if(tmp.indexOf('=') >= 0)
			{
				control = new StringTokenizer(tmp, "=");
				parseControl(control.nextToken().trim(), control.nextToken().trim(), false);
			} else {
				parseTone(tmp);
			}
		}
	}

	private void parseControl(final String name, final String value, final boolean isDefault)
		throws ToneParseException
	{
		if(name.equals("o"))
		{
			try {
				int scale = Integer.parseInt(value);

				if(scale >= 4 && scale <= 7)
				{
					defaultScale = scale;

					output("Scale: " + scale);
				} else {
					throw new ToneParseException(Errors.ERROR_RTX_INVALID_SCALE_SETTING,
						"Invalid scale control statement");
				}
			} catch(final NumberFormatException nfe) {
				throw new ToneParseException(Errors.ERROR_RTX_INVALID_SCALE_SETTING,
					"Invalid scale control statement");
			}
		} else if(name.equals("d")) {
			try {
				int duration = Integer.parseInt(value);

				if(duration == 1 || duration == 2 || duration == 4 || duration == 8 ||
					duration == 16 || duration == 32)
				{
					defaultDuration = duration;

					output("Duration: " + duration);
				} else {
					throw new ToneParseException(Errors.ERROR_RTX_INVALID_DURATION_SETTING,
						"Invalid duration control statement");
				}
			} catch(final NumberFormatException nfe) {
				throw new ToneParseException(Errors.ERROR_RTX_INVALID_DURATION_SETTING,
					"Invalid duration control statement");
			}
		} else if(name.equals("b")) {
			try {
				int bpm = Integer.parseInt(value);

				if(bpm == 25 || bpm == 28 || bpm == 31 || bpm == 35 || bpm == 40 || bpm == 45 ||
					bpm == 50 || bpm == 56 || bpm == 63 || bpm == 70 || bpm == 80 || bpm == 90 ||
					bpm == 100 || bpm == 112 || bpm == 125 || bpm == 140 || bpm == 160 || bpm == 180 ||
					bpm == 200 || bpm == 225 || bpm == 250 || bpm == 285 || bpm == 320 || bpm == 355 ||
					bpm == 400 || bpm == 450 || bpm == 500 || bpm == 565 || bpm == 635 || bpm == 715 ||
					bpm == 800 || bpm == 900)
				{
					defaultBPM = bpm;
				} else {
					throw new ToneParseException(Errors.ERROR_RTX_INVALID_BPM_SETTING,
						"Invalid beat-per-minute control statement");
				}

				output("Beats per minute: " + value);
			} catch(final NumberFormatException nfe) {
				throw new ToneParseException(Errors.ERROR_RTX_INVALID_BPM_SETTING,
					"Invalid beat-per-minute control statement");
			}
		} else if(name.equals("s")) {
			if(value.equals("s"))
			{
				output("Style: Stacatto");
			} else if(value.equals("n")) {
				output("Style: Natural");
			} else if(value.equals("c")) {
				output("Style: Continuous");
			} else {
				throw new ToneParseException(Errors.ERROR_RTX_INVALID_STYLE_SETTING,
					"Invalid style control setting");
			}

			// MobileStreams ignores style
		} else if(name.equals("l")) {
			try {
				int loop = Integer.parseInt(value);

				if(loop == 0)
				{
					output("No looping");
				} else if(loop > 0 && loop < 15) {
					output("Loop " + loop + " times");
				} else if(loop == 15) {
					output("Infite looping");
				} else {
					throw new ToneParseException(Errors.ERROR_RTX_INVALID_LOOP_SETTING,
						"Invalid loop control statement");
				}

				controlLoop = loop;
			} catch(final NumberFormatException nfe) {
				throw new ToneParseException(Errors.ERROR_RTX_INVALID_LOOP_SETTING,
					"Invalid loop control statement");
			}
		} else {
			output("Unknown control statement: " + name + " = " + value);
/* we're supposed to ignore unknown control statements according to the RTX spec */
		}
	}

	private void parseControlSection(final String str)
		throws ToneParseException
	{
		StringTokenizer controlPairs = new StringTokenizer(str, ","), controlParts = null;
		String tmp = null;

		while(controlPairs.hasMoreTokens())
		{
			tmp = controlPairs.nextToken().trim();
			controlParts = new StringTokenizer(tmp, "=");

			if(controlParts.countTokens() != 2)
			{
				throw new ToneParseException(Errors.ERROR_RTX_BAD_CONTROL_FORMAT,
					"Invalid format for control statement");
			}

			parseControl(controlParts.nextToken().trim(), controlParts.nextToken().trim(), true);
		}
	}

	private class Note
	{
		public int scale, duration, bpm;
		public String note;
		public boolean dotted;
	}

	private static String getBits(final int i, final int j)
	{
		String s = Integer.toBinaryString(i);
		int k = s.length();

		if(j == k)
			return s;

		if(j < k)
			return s.substring(s.length() - j);

		for(int l = k; l < j; l++)
			s = "0" + s;

		return s;
	}

	private String getDurationString(final int scale)
		throws ToneParseException
	{
		String value = (String)OTTConstants.durationMap.get("" + scale);

		if(value == null)
		{
			throw new ToneParseException(Errors.ERROR_RTX_INVALID_DURATION_SETTING,
				"Invalid duration setting");
		}

		return value;
	}

	private String getScaleString(final int scale)
		throws ToneParseException
	{
		String value = (String)OTTConstants.scaleMap.get("" + scale);

		if(value == null)
		{
			throw new ToneParseException(Errors.ERROR_RTX_INVALID_SCALE_SETTING,
				"Invalid scale setting");
		}

		return value;
	}

	private String getTempoString(final int bpm)
		throws ToneParseException
	{
		String value = (String)OTTConstants.tempoMap.get("" + bpm);

		if(value == null)
		{
			throw new ToneParseException(Errors.ERROR_RTX_INVALID_BPM_SETTING,
				"Invalid tempo setting");
		}

		return value;
	}

	public class ToneParseException extends Exception
	{
		private int error = -1;

		public ToneParseException(final int err, final String message)
		{
			super(message);
			setErrorNumber(err);
		}

		public void setErrorNumber(final int err)
		{
			error = err;
		}

		public int getErrorNumber()
		{
			return error;
		}
	}
}
